"""
A complete bitboard representation of a chessboard, with all the methods needed
to play a game of chess.
"""


try:
    import importlib.resources as pkg_resources
except ImportError:
    # Python < 3.7
    pkg_resources = None
import random
import sys
from copy import copy
from math import log2
from typing import Tuple, Iterable

from chessengine.moves import (
    get_white_pawn_moves,
    get_white_rook_moves,
    get_white_bishop_moves,
    get_white_knight_moves,
    get_white_king_moves,
    get_white_queen_moves,
    get_black_pawn_moves,
    get_black_rook_moves,
    get_black_bishop_moves,
    get_black_knight_moves,
    get_black_king_moves,
    get_black_queen_moves,
)
from chessengine.lookup_tables import (
    mask_position,
    clear_position,
    coords_to_pos,
    pos_to_coords,
    san_piece_map,
)
from chessengine.utils import (
    get_bit_positions,
    get_file,
    get_rank,
    clear_lines,
    get_input,
    change_turn,
)
from chessengine.pgn.parser import PGNParser, SAN_MOVE_REGEX


class Board:
    """
    A class implementing a bitboard representation of a chess board.
    A particular bitboard can be accessed via the get_bitboard method or
    as an attribute with the name <side>_<piece>s. For example, ``white_pawns``,
    ``black_bishops``, ``white_queens``, ``black_kings``.

    :param side: The side that the _board_ will play. Should be one of "white" or "black"
    """

    def __init__(self, side: str):
        self.white_pawns = 65280  # (A2 to H2)
        self.white_rooks = 129  # (A1 and H1)
        self.white_knights = 66  # (B1 and G1)
        self.white_bishops = 36  # (C1 and F1)
        self.white_queens = 8  # (D1)
        self.white_kings = 16  # (E1)

        self.black_pawns = 71776119061217280  # (A7 to H7)
        self.black_rooks = 9295429630892703744  # (A8 and H8)
        self.black_knights = 4755801206503243776  # (B8 and G8)
        self.black_bishops = 2594073385365405696  # (C8 and F8)
        self.black_queens = 576460752303423488  # (D8)
        self.black_kings = 1152921504606846976  # (E8)

        self.all_white = (
            self.white_pawns
            | self.white_rooks
            | self.white_knights
            | self.white_bishops
            | self.white_queens
            | self.white_kings
        )

        self.all_black = (
            self.black_pawns
            | self.black_rooks
            | self.black_knights
            | self.black_bishops
            | self.black_queens
            | self.black_kings
        )

        self.all_pieces = self.all_black | self.all_white

        self.piece_count = {
            ("white", "kings"): 1,
            ("white", "queens"): 1,
            ("white", "rooks"): 2,
            ("white", "bishops"): 2,
            ("white", "knights"): 2,
            ("white", "pawns"): 8,
            ("black", "kings"): 1,
            ("black", "queens"): 1,
            ("black", "rooks"): 2,
            ("black", "bishops"): 2,
            ("black", "knights"): 2,
            ("black", "pawns"): 8,
        }

        if side.lower().strip() not in ["black", "white"]:
            raise ValueError(f'side must be one of "black" or "white". Got {side}')
        self.side = side.lower().strip()
        self.opponent_side = "black" if self.side == "white" else "white"
        self.en_passant_position = 0

        # A dictionary matching a side and piece to its corresponding bit board.
        # Useful when we want to iterate through all of the bitboards of the board.
        self.board = {
            ("white", "kings"): self.white_kings,
            ("white", "queens"): self.white_queens,
            ("white", "rooks"): self.white_rooks,
            ("white", "bishops"): self.white_bishops,
            ("white", "knights"): self.white_knights,
            ("white", "pawns"): self.white_pawns,
            ("black", "kings"): self.black_kings,
            ("black", "queens"): self.black_queens,
            ("black", "rooks"): self.black_rooks,
            ("black", "bishops"): self.black_bishops,
            ("black", "knights"): self.black_knights,
            ("black", "pawns"): self.black_pawns,
        }

        # Keep track of all moves made
        self.moves = []

    def __repr__(self):
        piece_list = ["\u2001" for _ in range(64)]
        unicode_piece = {
            ("white", "kings"): "\u2654",
            ("white", "queens"): "\u2655",
            ("white", "rooks"): "\u2656",
            ("white", "bishops"): "\u2657",
            ("white", "knights"): "\u2658",
            ("white", "pawns"): "\u2659",
            ("black", "kings"): "\u265A",
            ("black", "queens"): "\u265B",
            ("black", "rooks"): "\u265C",
            ("black", "bishops"): "\u265D",
            ("black", "knights"): "\u265E",
            ("black", "pawns"): "\u265F",
        }
        ranks = ["8", "7", "6", "5", "4", "3", "2", "1"]
        files = ["\u2001", "a", "b", "c", "d", "e", "f", "g", "h", "\u2001"]

        def add_bitboard_to_repr(board, s, p):
            board_string = bin(board)[2:]
            board_string = "0" * (64 - len(board_string)) + board_string
            for _ in range(64):
                if board_string[_] == "1":
                    piece_list[_] = unicode_piece[(s, p)]

        for side, piece in self.board:
            add_bitboard_to_repr(self.board[(side, piece)], side, piece)

        r = range(8)
        if self.side == "white":
            r = reversed(r)
            files = list(reversed(files))

        board_repr = ""
        board_repr += "\u2001" + "\u2001".join(files) + "\n"

        for i in r:
            board_repr += ranks[i] + "\u2001\u2001"

            if self.side == "white":
                board_repr += "\u2001".join(piece_list[8 * i : 8 * i + 8])
            else:
                board_repr += "\u2001".join(piece_list[8 * i : 8 * i + 8][::-1])

            board_repr += "\u2001\u2001" + ranks[i]
            board_repr += "\n"

        board_repr += "\u2001" + "\u2001".join(files) + "\n"
        return board_repr

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if self.side != other.side:
            return False
        return str(self) == str(other)

    def __hash__(self):
        hash_str = f"{self.side[0]} "
        for side, piece in [
            ("white", "kings"),
            ("white", "queens"),
            ("white", "rooks"),
            ("white", "bishops"),
            ("white", "knights"),
            ("white", "pawns"),
            ("black", "kings"),
            ("black", "queens"),
            ("black", "rooks"),
            ("black", "bishops"),
            ("black", "knights"),
            ("black", "pawns"),
        ]:
            hash_str += str(self.board[(side, piece)]) + " "
        return hash(hash_str)

    @property
    def score(self):
        """
        The "score" of the board. A higher/more positive score favors white,
        a lower/more negative score favors black.
        """
        K = self.piece_count[("white", "kings")]
        Q = self.piece_count[("white", "queens")]
        R = self.piece_count[("white", "rooks")]
        B = self.piece_count[("white", "bishops")]
        N = self.piece_count[("white", "knights")]
        P = self.piece_count[("white", "pawns")]
        k = self.piece_count[("black", "kings")]
        q = self.piece_count[("black", "queens")]
        r = self.piece_count[("black", "rooks")]
        b = self.piece_count[("black", "bishops")]
        n = self.piece_count[("black", "knights")]
        p = self.piece_count[("black", "pawns")]
        s = 200 * (K - k) + 9 * (Q - q) + 5 * (R - r) + 3 * (B - b + N - n) + (P - p)
        if self.side == "white":
            return s
        return -s

    @property
    def board_pieces(self):
        """
        A list of all bitboards of the pieces that belong to the Board.
        """
        if self.side == "white":
            return [
                ("white", "kings"),
                ("white", "queens"),
                ("white", "rooks"),
                ("white", "bishops"),
                ("white", "knights"),
                ("white", "pawns"),
            ]
        return [
            ("black", "kings"),
            ("black", "queens"),
            ("black", "rooks"),
            ("black", "bishops"),
            ("black", "knights"),
            ("black", "pawns"),
        ]

    @property
    def opponent_pieces(self):
        """
        A list of all bitboards of the pieces that belong to the opponent.
        """
        if self.side == "black":
            return [
                ("white", "kings"),
                ("white", "queens"),
                ("white", "rooks"),
                ("white", "bishops"),
                ("white", "knights"),
                ("white", "pawns"),
            ]
        return [
            ("black", "kings"),
            ("black", "queens"),
            ("black", "rooks"),
            ("black", "bishops"),
            ("black", "knights"),
            ("black", "pawns"),
        ]

    def copy(self):
        """
        Create and return a copy of the board.
        """
        return copy(self)

    def get_side_bitboard(self, side: str) -> int:
        """
        Returns the bitboard containing all pieces for the given side.

        :param side: "white" or "black"
        :return: If ``side == "white"``, returns ``self.all_white``, else ``self.all_black``
        """
        if side == "white":
            return self.all_white
        return self.all_black

    def get_bitboard(self, side: str, piece: str) -> int:
        """
        Returns the bitboard of the passed side for the passed pieces.
        For example, calling with side="black" and piece="king" will return the black_kings bitboard, and so on.

        Raises AttributeError if a bitboard with an invalid name is requested. See above for the bitboard naming
        convention.

        :param side: "white" or "black"
        :param piece: Can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
        :return: Bitboard
        """
        attrname = side + "_" + piece
        return getattr(self, attrname)

    def get_self_piece_bitboard(self, piece: str) -> int:
        """
        Returns the bitboard corresponding to the passed piece, considering the board's
        own side. i.e. - If the board is white, calling with piece="king" will return
        white king, etc.
        :param piece: Can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
        :return: Bitboard
        """
        return self.get_bitboard(side=self.side, piece=piece)

    def update_board_state(self) -> None:
        """
        Updates all Board attributes when a move is made or the state of the Board
        changes in any way. You should call this if you manually make any changes to
        bitboards attributes, otherwise it is called automatically.
        """
        self.all_white = (
            self.white_pawns
            | self.white_rooks
            | self.white_knights
            | self.white_bishops
            | self.white_queens
            | self.white_kings
        )

        self.all_black = (
            self.black_pawns
            | self.black_rooks
            | self.black_knights
            | self.black_bishops
            | self.black_queens
            | self.black_kings
        )

        self.all_pieces = self.all_black | self.all_white

        self.board = {
            ("white", "kings"): self.white_kings,
            ("white", "queens"): self.white_queens,
            ("white", "rooks"): self.white_rooks,
            ("white", "bishops"): self.white_bishops,
            ("white", "knights"): self.white_knights,
            ("white", "pawns"): self.white_pawns,
            ("black", "kings"): self.black_kings,
            ("black", "queens"): self.black_queens,
            ("black", "rooks"): self.black_rooks,
            ("black", "bishops"): self.black_bishops,
            ("black", "knights"): self.black_knights,
            ("black", "pawns"): self.black_pawns,
        }

    def set_bitboard(self, side: str, piece: str, board: int) -> None:
        """
        Sets the bitboard for the passed arguments to the passed bitboard

        :param side: "white" or "black"
        :param piece: Can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
        :param board: The bitboard to be set
        """
        attrname = side + "_" + piece
        setattr(self, attrname, board)
        self.update_board_state()

    def identify_piece_at(self, position: int) -> tuple:
        """
        Identifies if there is any piece on the position passed.

        :param position: A power of 2 corresponding to a position on the board. See :ref:`position_representation`
        :return: Returns a 3-tuple of the format (side, piece, bitboard) where side is the side of
            the piece identified at position (e.g, "black"), piece is the type of piece identified
            at position (e.g, "bishops"), and bitboard is the bitboard of the piece (e.g, Board.black_bishops).
        """
        for side, piece in self.board:
            board = self.board[(side, piece)]
            if board & position > 0:
                return side, piece, board
        return None, None, None

    def move(self, start: int, end: int, track: bool = True) -> None:
        """
        Moves the piece at start to end. Doesn't check anything, just makes
        the move (unless the start or end positions are invalid).

        :param start: The start position of the move. See :ref:`position_representation`
        :param end: The end position of the move. See :ref:`position_representation`
        :param track: If ``True``, the move made will be stored in self.moves
        """
        if not 1 <= start <= 2**63:
            raise ValueError(
                f"The start position is outside the board - moving from {log2(start)} to {log2(end)}"
            )
        if not 1 <= end <= 2**63:
            raise ValueError(
                f"The end position is outside the board - moving from {log2(start)} to {log2(end)}"
            )

        start_side, start_piece, start_board = self.identify_piece_at(start)
        if start_side is None:
            raise ValueError(f"There is no piece at position {log2(start)} to move")

        end_side, end_piece, end_board = self.identify_piece_at(end)
        if end_side == start_side:
            raise ValueError(
                f"Can't move from {log2(start)} to {log2(end)}, both positions have {end_side} pieces."
            )

        if track:
            # Keep track of the board state before the move was made so we can undo
            start_state = (start, end, end_side, end_piece, end_board)
            self.moves.append(start_state)

        # Check en passant moves
        if start_piece == "pawns":
            if start_side == "white":
                # Check en passant status
                if get_rank(end) - get_rank(start) == 2:
                    self.en_passant_position = start << 8

                # Check if a pawn captured by an en passant move
                elif get_file(start) != get_file(end):
                    # White pawn made an en passant move
                    black_pawn_bb = self.get_bitboard("black", "pawns")
                    black_pawn_bb &= clear_position[end >> 8]
                    self.set_bitboard("black", "pawns", black_pawn_bb)
                    self.piece_count[("black", "pawns")] -= 1
                    self.en_passant_position = 0

                # Clear self.en_passant_position
                else:
                    self.en_passant_position = 0

            else:
                # Check en passant status
                if get_rank(start) - get_rank(end) == 2:
                    self.en_passant_position = start >> 8

                # Check if a pawn captured by an en passant move
                elif get_file(start) != get_file(end):
                    # Black pawn made an en passant move
                    white_pawn_bb = self.get_bitboard("white", "pawns")
                    white_pawn_bb &= clear_position[end << 8]
                    self.set_bitboard("white", "pawns", white_pawn_bb)
                    self.piece_count[("white", "pawns")] -= 1
                    self.en_passant_position = 0

                # Clear self.en_passant_position
                else:
                    self.en_passant_position = 0

        else:
            self.en_passant_position = 0

        if end_piece is not None:
            # Clear the captured piece's position (set "end" to 0)
            opp_side_board = self.get_bitboard(end_side, end_piece)
            opp_side_board &= clear_position[end]
            self.set_bitboard(end_side, end_piece, opp_side_board)
            self.piece_count[(end_side, end_piece)] -= 1

        # Clear the moved piece's original position (set "start" to 0)
        move_side_board = self.get_bitboard(start_side, start_piece)
        move_side_board &= clear_position[start]

        # Set the moved piece's final position (set "end" to 1)
        move_side_board |= mask_position[end]
        self.set_bitboard(start_side, start_piece, move_side_board)

    def move_raw(self, start: int, end: int, track: bool = True) -> None:
        """
        Moves the piece at start to end. Checks if the move is a valid move
        to make given the current state of the board.

        :param start: The start position of the move. See :ref:`position_representation`
        :param end: The end position of the move. See :ref:`position_representation`
        :param track: If ``True``, the move made will be stored in self.moves
        """
        side, piece, board = self.identify_piece_at(start)
        if side is None:
            raise ValueError(
                f"There is no piece at {pos_to_coords[int(log2(start))]} to move."
            )
        moves = self.get_moves(side=side, piece=piece)
        if (start, end) not in moves:
            raise ValueError(
                f"{pos_to_coords[int(log2(start))]} to {pos_to_coords[int(log2(end))]} is not a valid move for {side}"
            )
        self.move(start=start, end=end, track=track)

    def move_san(self, move: str, side: str) -> None:
        """
        Make a move given in standard algebraic notation.

        :param move: A move given in SAN
        :param side: "white" or "black"
        """
        # TODO - Add support for undoing castling
        if "0-0-0" in move:
            # queen side castle
            if side == "white":
                self.move(1, 2**3, track=False)
                self.move(2**4, 2**2)
            else:
                self.move(2**56, 2**59, track=False)
                self.move(2**60, 2**58)
        elif "0-0" in move:
            # king side castle
            if side == "white":
                self.move(2**7, 2**5, track=False)
                self.move(2**4, 2**6)
            else:
                self.move(2**63, 2**61, track=False)
                self.move(2**60, 2**62)
        else:
            # regular move
            match = SAN_MOVE_REGEX.match(move)
            if match is None:
                raise ValueError(f'Couldn\'t parse move "{move}".')

            groups = match.groups()
            if groups[0] is None:
                piece_moved = san_piece_map["P"]
            else:
                piece_moved = san_piece_map[groups[0].upper()]

            end_pos = 2 ** coords_to_pos[groups[3].upper()]
            moves = self.get_moves(side, piece_moved)
            if groups[1] is None:
                # No file provided in the SAN
                for m in moves:
                    if m[1] == end_pos:
                        self.move(start=m[0], end=m[1])
                        break
                else:
                    raise ValueError(f"{move} is not a valid move for {side}.")
            else:
                if groups[2] is None:
                    # No rank provided in the SAN
                    for m in moves:
                        file = get_file(m[0])
                        if (
                            groups[1].upper() == "ABCDEFGH"[file - 1]
                            and m[1] == end_pos
                        ):
                            self.move(start=m[0], end=m[1])
                            break
                    else:
                        raise ValueError(f"{move} is not a valid move for {side}.")
                else:
                    # File and rank both present in the SAN
                    start_pos = 2 ** coords_to_pos[groups[1].upper() + groups[2]]
                    if (start_pos, end_pos) not in moves:
                        raise ValueError(f"{move} is not a valid move for {side}.")
                    self.move(start=start_pos, end=end_pos)

    def make_moves(self, *moves: Iterable[tuple[int, int]]) -> None:
        """
        Given a number of moves as tuples (start, end), call
        Board.move on all. Tracks all moves by default in ``self.moves``
        """
        for start, end in moves:
            self.move(start, end)

    def undo_move(self) -> None:
        """
        Undo the last tracked move.
        """
        if not self.moves:
            raise RuntimeError("No moves have been made yet to undo.")
        end, start, side, piece, board = self.moves.pop()
        self.move(start=start, end=end, track=False)
        if side is not None:
            self.set_bitboard(side, piece, board)
            self.piece_count[(side, piece)] += 1

    def get_moves(
        self, side: str, piece: str = None, position: int = None
    ) -> list[tuple[int, int]]:
        """
        Get all end positions a piece of side can reach starting from position.
        ``side`` is always required, piece and position are optional.

        If piece is not specified, gets all moves for all pieces of the passed side, i.e. get
        all valid moves for white or black.

        If side and piece are specified and position is not, gets all valid moves for the
        specified side and piece on the board, i.e. if side is "white" and piece is "rooks",
        gets all valid moves for all white rooks on the board.

        If side, piece, and position are specified, gets all moves for the piece present on position.

        :param side: "white" or "black"
        :param piece: Can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
        :param position: A power of 2 corresponding to a position on the board. See :ref:`position_representation`
        :return: A list of moves as tuples ``(start, end)``, where ``start`` and ``end`` are positions on the board.
            See :ref:`position_representation`
        """
        if piece is not None:
            if position is None:
                moves = []
                positions = get_bit_positions(self.get_bitboard(side, piece))
                for position in positions:
                    moves.extend(self.get_moves(side, piece, position))
                return moves
            else:
                move_gens = {
                    ("white", "kings"): get_white_king_moves,
                    ("white", "queens"): get_white_queen_moves,
                    ("white", "rooks"): get_white_rook_moves,
                    ("white", "bishops"): get_white_bishop_moves,
                    ("white", "knights"): get_white_knight_moves,
                    ("white", "pawns"): get_white_pawn_moves,
                    ("black", "kings"): get_black_king_moves,
                    ("black", "queens"): get_black_queen_moves,
                    ("black", "rooks"): get_black_rook_moves,
                    ("black", "bishops"): get_black_bishop_moves,
                    ("black", "knights"): get_black_knight_moves,
                    ("black", "pawns"): get_black_pawn_moves,
                }
                # TODO - Add support for en passant move detection
                return move_gens[(side, piece)](self, position)
        else:
            moves = []
            if side == self.side:
                pieces = self.board_pieces
            else:
                pieces = self.opponent_pieces
            for side, piece in pieces:
                positions = get_bit_positions(self.get_bitboard(side, piece))
                for position in positions:
                    moves.extend(self.get_moves(side, piece, position))
            return moves

    def search_forward(self, depth: int = 4) -> tuple[int, tuple[int, int]]:
        """
        Execute an alpha-beta pruned depth-first search to find the optimal move from
        the current board state.

        :param depth: int - The number of plies to search (1 move is 2 plies). Default = 4 plies.
        :return: A 2-tuple where the first element is the best board score found, and the second
            element is the best found move as a 2 tuple containing the start position and the end position
        """
        maximize = self.side == "white"
        best_score = -1000 if maximize else 1000

        moves = self.get_moves(self.side)
        best_move = moves[0]

        for move in moves:
            self.move(start=move[0], end=move[1])
            value = self.alpha_beta_search(
                depth=depth - 1, maximizing_player=not maximize
            )
            self.undo_move()

            if maximize and value >= best_score:
                best_score = value
                best_move = move
            elif not maximize and value <= best_score:
                best_score = value
                best_move = move
        return best_score, best_move

    def alpha_beta_search(
        self,
        depth: int = 4,
        alpha: int = -1000,
        beta: int = 1000,
        maximizing_player: bool = True,
    ) -> int:
        """
        Execute an alpha-beta pruned search. You probably won't need to
        call this function yourself, use Board.search_forward instead.

        :param depth: The number of plies to search forward (default=4)
        :param alpha: The minimum score that the maximizing player is guaranteed (default=-1000). You probably won't need to specify this argument.
        :param beta: The maximum score that the minimizing player is guaranteed (default=1000). You probably won't need to specify this argument.
        :param maximizing_player: True if white is searching for a move, False if black is searching for a move.

        :return: The score of the best board position found.
        """
        if depth == 0:
            return self.score

        if maximizing_player:
            value = -1000
            moves = self.get_moves(self.side)
            for move in moves:
                self.move(start=move[0], end=move[1])
                final_score = self.alpha_beta_search(depth - 1, alpha, beta, False)
                value = max(value, final_score)
                self.undo_move()
                if value >= beta:
                    break
                alpha = max(alpha, value)
            return value
        else:
            value = 1000
            moves = self.get_moves(self.opponent_side)
            for move in moves:
                self.move(start=move[0], end=move[1])
                final_score = self.alpha_beta_search(depth - 1, alpha, beta, True)
                value = min(value, final_score)
                self.undo_move()
                if value <= alpha:
                    break
                beta = min(beta, value)
            return value

    def handle_player_move(
        self, side_to_move: str, last_move: str
    ) -> Tuple[str, int, bool]:
        """
        Ask for user input until accepted. When a valid input is received, execute the move.

        :param side_to_move: "white" or "black"
        :param last_move: The last move made by the board to print out to the user
        :return: A 3-tuple of the form - the valid (move, lines_printed, move_undone)
            where move is the valid move received as input, lines_printed is the number
            of lines printed, and move_undone is a boolean that is ``True`` when the user
            chose to undo their previous move
        """
        move = ""
        lines_added = 0
        input_accepted = False

        while not input_accepted:
            print(last_move)
            move = get_input(
                f"Enter your move in standard algebraic notation ({side_to_move.capitalize()}'s turn) - "
            )
            lines_added += 2
            if move.lower() == "q":
                print("Thanks for playing!")
                sys.exit(0)
            if move.lower() == "u":
                try:
                    self.undo_move()
                    self.undo_move()  # Undo both player's moves
                except RuntimeError:
                    print("No moves have been made yet to undo!")
                    lines_added += 1
                return move, lines_added, True
            # Input was normal move
            try:
                if " to " in move:
                    # Input was not in SAN
                    _ = move.upper().strip().split()
                    start = 2 ** coords_to_pos[_[0]]
                    end = 2 ** coords_to_pos[_[-1]]
                    self.move_raw(start=start, end=end)
                else:
                    # Input was in SAN
                    self.move_san(move=move, side=side_to_move)
                input_accepted = True
                last_move = f"{side_to_move.capitalize()} moved {move}"
            except ValueError as e:
                print(e)
                lines_added += 1
            except KeyError as e:
                print(f"{e} is not a valid square. Please try again.")
                lines_added += 1
        return move, lines_added, False

    def play(self, search_depth: int = 4) -> None:
        """
        Play a game of chess against the computer.

        :param search_depth: The number of plies the computer should search forward. Be careful
            passing values about 4 as the search depth. It increases the running time of the
            move search exponentially.
        """
        parser = None
        if pkg_resources is not None:
            loading_messages = [
                "Searching for opening moves.",
                "Reading an opening book.",
                "Waking up.",
                "Reading thousands of past games.",
            ]

            parser = PGNParser()
            print(random.choice(loading_messages))
            opening_files = pkg_resources.contents("chessengine.openings")
            for child in opening_files:
                file_path = pkg_resources.path("chessengine.openings", child)
                with file_path as f:
                    if f.suffix == ".pgn":
                        print(".")
                        parser.parse(f)
            print(f"\nRead through {len(parser.games)} games.")
        print(f"Set search depth to {search_depth}")

        print(self)
        side_to_move = "white"
        in_game_tree = parser is not None
        current_node = parser.root_node if in_game_tree else None
        lines_printed = 11
        last_move = ""
        while True:
            clear_lines(lines_printed)
            print(self)
            lines_printed = 11
            if side_to_move == self.side:
                if in_game_tree:
                    move, node = random.choice(list(current_node.children.items()))
                    self.move_san(move=move, side=side_to_move)
                    current_node = node
                    last_move = f"Board moves {move}"
                else:
                    best_score, best_move = self.search_forward(search_depth)
                    self.move(best_move[0], best_move[1])
                    last_move = f"Board moves from {pos_to_coords[log2(best_move[0])]} to {pos_to_coords[log2(best_move[1])]}"
            else:
                move, lines_added, move_undone = self.handle_player_move(
                    side_to_move, last_move
                )
                lines_printed += lines_added
                last_move = f"{side_to_move.capitalize()} moved {move}"

                if (
                    move_undone
                ):  # return to outer loop, so both sides need to make a new move
                    continue

                if in_game_tree:
                    try:
                        current_node = current_node.get_child(move)
                    except ValueError:
                        in_game_tree = False
                print(f"You moved {move}")
                lines_printed += 1

            side_to_move = change_turn(side_to_move)

    def play_pvp(self) -> None:
        """
        Play a game of chess on the terminal with another player
        """
        print(self)
        lines_printed = 11

        side_to_move = "white"
        last_move = ""
        while True:
            clear_lines(lines_printed)
            print(self)
            lines_printed = 11

            move, lines_added, move_undone = self.handle_player_move(
                side_to_move, last_move
            )
            lines_printed += lines_added
            last_move = f"{side_to_move.capitalize()} moved {move}"

            if (
                move_undone
            ):  # return to outer loop, so both sides need to make a new move
                continue

            side_to_move = change_turn(side_to_move)



"""
Functions for generating moves for all supported pieces in all supported scenarios
"""


from chessengine.utils import get_rank, get_file

HIGHEST_SQUARE = 2**63


def check_valid_position(
    board,
    side: str,
    start: int,
    end: int,
    moves: list[tuple[int, int]],
    auto_add: bool = True,
) -> tuple[bool, bool]:
    """
    Returns a tuple(bool). The first value indicates whether
    this position is valid to move to, the second indicates
    whether this is the last valid position in the direction
    that the piece is currently moving (whether the piece
    should try to explore the next position in the direction
    that is being explored)

    If auto_add is True, adds the position to the moves list
    if the move is valid
    """
    if end > HIGHEST_SQUARE:
        return False, True
    if end <= 0:
        return False, True
    if end & board.all_pieces == 0:
        if auto_add:
            moves.append((start, end))
        return True, False
    elif side == "white" and end & board.all_white == 0:
        if auto_add:
            moves.append((start, end))
        return True, True
    elif side == "black" and end & board.all_black == 0:
        if auto_add:
            moves.append((start, end))
        return True, True
    else:
        return False, True


def get_rook_moves(board, side: str, position: int) -> list[int]:
    """
    Returns a list of end positions a rook of side=side can reach starting at position

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param side: The side of the rook. Either "white" or "black"
    :param position: The position the rook starts on. See :ref:`position_representation`
    """
    moves = []

    _ = position
    while True:
        # Move rank up
        _ = _ << 8
        valid, should_break = check_valid_position(board, side, position, _, moves)
        if should_break:
            break

    _ = position
    while True:
        # Move rank down
        _ = _ >> 8
        valid, should_break = check_valid_position(board, side, position, _, moves)
        if should_break:
            break

    file = get_file(position)
    max_right = 8 - file
    _ = position
    for i in range(max_right):
        # Move right
        _ = _ << 1
        valid, should_break = check_valid_position(board, side, position, _, moves)
        if should_break:
            break

    max_left = file - 1
    _ = position
    for i in range(max_left):
        # Move left
        _ = _ >> 1
        valid, should_break = check_valid_position(board, side, position, _, moves)
        if should_break:
            break

    return moves


def get_white_rook_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a white rook starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the rook starts on. See :ref:`position_representation`
    """
    return get_rook_moves(board, "white", position)


def get_black_rook_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a black rook starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the rook starts on. See :ref:`position_representation`
    """
    return get_rook_moves(board, "black", position)


def get_bishop_moves(board, side: str, position: int) -> list[int]:
    """
    Returns a list of end positions a bishop of side=side can reach starting at position

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param side: The side of the bishop. "white" or "black"
    :param position: The position the bishop starts on. See :ref:`position_representation`
    """
    moves = []
    file = get_file(position)
    max_right = 8 - file
    _ = position
    for i in range(max_right):
        _ = _ << 9
        valid, should_break = check_valid_position(board, side, position, _, moves)
        if should_break:
            break

    _ = position
    for i in range(max_right):
        _ = _ >> 7
        valid, should_break = check_valid_position(board, side, position, _, moves)
        if should_break:
            break

    max_left = file - 1
    _ = position
    for i in range(max_left):
        _ = _ << 7
        valid, should_break = check_valid_position(board, side, position, _, moves)
        if should_break:
            break

    _ = position
    for i in range(max_left):
        _ = _ >> 9
        valid, should_break = check_valid_position(board, side, position, _, moves)
        if should_break:
            break

    return moves


def get_white_bishop_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a white bishop starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the bishop starts on. See :ref:`position_representation`
    """
    return get_bishop_moves(board, "white", position)


def get_black_bishop_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a black bishop starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the bishop starts on. See :ref:`position_representation`
    """
    return get_bishop_moves(board, "black", position)


def get_knight_moves(board, side: str, position: int) -> list[int]:
    """
    Returns a list of end positions a knight starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param side: The side of the knight. "white" or "black"
    :param position: The position the knight starts on. See :ref:`position_representation`
    """
    moves = []
    rank = get_rank(position)
    file = get_file(position)

    if rank >= 3:
        if file >= 2:
            _ = position >> 17
            check_valid_position(board, side, position, _, moves)
        if file <= 7:
            _ = position >> 15
            check_valid_position(board, side, position, _, moves)

    if rank >= 2:
        if file >= 3:
            _ = position >> 10
            check_valid_position(board, side, position, _, moves)
        if file <= 6:
            _ = position >> 6
            check_valid_position(board, side, position, _, moves)

    if rank <= 6:
        if file >= 2:
            _ = position << 15
            check_valid_position(board, side, position, _, moves)
        if file <= 7:
            _ = position << 17
            check_valid_position(board, side, position, _, moves)

    if rank <= 7:
        if file >= 3:
            _ = position << 6
            check_valid_position(board, side, position, _, moves)
        if file <= 6:
            _ = position << 10
            check_valid_position(board, side, position, _, moves)

    return moves


def get_white_knight_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a white knight starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the knight starts on. See :ref:`position_representation`
    """
    return get_knight_moves(board, "white", position)


def get_black_knight_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a black knight starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the knight starts on. See :ref:`position_representation`
    """
    return get_knight_moves(board, "black", position)


def get_king_moves(board, side: str, position: int) -> list[int]:
    """
    Returns a list of end positions a king starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param side: The side of the king. "white" or "black".
    :param position: The position the king starts on. See :ref:`position_representation`
    """
    moves = []

    rank = get_rank(position)
    file = get_file(position)

    if rank >= 2:
        _ = position >> 8
        check_valid_position(board, side, position, _, moves)

        if file >= 2:
            _ = position >> 9
            check_valid_position(board, side, position, _, moves)

        if file <= 7:
            _ = position >> 7
            check_valid_position(board, side, position, _, moves)

    if rank <= 7:
        _ = position << 8
        check_valid_position(board, side, position, _, moves)

        if file >= 2:
            _ = position << 7
            check_valid_position(board, side, position, _, moves)

        if file <= 7:
            _ = position << 9
            check_valid_position(board, side, position, _, moves)

    if file >= 2:
        _ = position >> 1
        check_valid_position(board, side, position, _, moves)

    if file <= 7:
        _ = position << 1
        check_valid_position(board, side, position, _, moves)
    return moves


def get_white_king_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a white king starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the king starts on. See :ref:`position_representation`
    """
    return get_king_moves(board, "white", position)


def get_black_king_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a black king starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the king starts on. See :ref:`position_representation`
    """
    return get_king_moves(board, "black", position)


def get_white_queen_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a white queen starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the queen starts on. See :ref:`position_representation`
    """
    return get_white_rook_moves(board, position) + get_white_bishop_moves(
        board, position
    )


def get_black_queen_moves(board, position: int) -> list[int]:
    """
    Returns a list of end positions a black queen starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the queen starts on. See :ref:`position_representation`
    """
    return get_black_rook_moves(board, position) + get_black_bishop_moves(
        board, position
    )


def get_white_pawn_moves(board, position: int) -> list[tuple[int, int]]:
    """
    Returns a list of end positions a white pawn starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the pawn starts on. See :ref:`position_representation`
    """
    rank = get_rank(position)
    if rank == 8:
        return []

    moves = []
    _ = position << 8
    if board.all_pieces & _ == 0:
        moves.append((position, _))
        if rank == 2:
            _ = position << 16
            if board.all_pieces & _ == 0:
                moves.append((position, _))
    file = get_file(position)
    if file >= 2:
        _ = position << 7
        if board.all_black & _ > 0:
            moves.append((position, _))
    if file <= 7:
        _ = position << 9
        if board.all_black & _ > 0:
            moves.append((position, _))

    en_passant_position = board.en_passant_position
    if en_passant_position == position << 7 and file >= 2:
        _ = position << 7
        check_valid_position(board, "white", position, _, moves)
    if en_passant_position == position << 9 and file <= 7:
        _ = position << 9
        check_valid_position(board, "white", position, _, moves)

    return moves


def get_black_pawn_moves(board, position: int) -> list[tuple[int, int]]:
    """
    Returns a list of end positions a black pawn starting at position can reach

    :param board: A :ref:`chessengine.bitboard.Board <Board>` object
    :param position: The position the pawn starts on. See :ref:`position_representation`
    """
    rank = get_rank(position)
    if rank == 1:
        return []

    moves = []
    _ = position >> 8
    if board.all_pieces & _ == 0:
        moves.append((position, _))
        if rank == 7:
            _ = position >> 16
            if board.all_pieces & _ == 0:
                moves.append((position, _))
    file = get_file(position)
    if file >= 2:
        _ = position >> 9
        if board.all_white & _ > 0:
            moves.append((position, _))
    if file <= 7:
        _ = position >> 7
        if board.all_white & _ > 0:
            moves.append((position, _))

    en_passant_position = board.en_passant_position
    if en_passant_position == position >> 9 and file >= 2:
        _ = position >> 9
        check_valid_position(board, "black", position, _, moves)
    if en_passant_position == position >> 7 and file <= 7:
        _ = position >> 7
        check_valid_position(board, "black", position, _, moves)

    return moves



from chessengine.utils import get_bit_positions, lsb_pos

import unittest


class TestUtils(unittest.TestCase):
    def test_get_bit_positions(self):
        self.assertEqual(get_bit_positions(0b110100), [0b100, 0b10000, 0b100000])
        self.assertEqual(get_bit_positions(0b110101), [0b1, 0b100, 0b10000, 0b100000])
        self.assertEqual(get_bit_positions(0b10), [0b10])
        self.assertEqual(get_bit_positions(0), [])

    def test_lsb_pos(self):
        test_cases = (
            (0b1010, 0b10),
            (0b1110010, 0b10),
            (0b1000010, 0b10),
            (0b1001, 0b1),
            (0b1, 0b1),
        )
        for arg, expected in test_cases:
            with self.subTest(arg=arg):
                self.assertEqual(lsb_pos(arg), expected)


if __name__ == "__main__":
    unittest.main()



"""
Defines the Game and GameNode class, used to parse PGN files and
create the opening book.
"""


class Game:
    """
    A class representing a game, keeping track of the moves made
    throughout the game and the GameNodes reached as a result.
    Used to record a game and replay it.
    """

    def __init__(self, root_node) -> None:
        self.headers: dict[str:str] = {}
        self.root_node: GameNode = root_node
        self.result = ""
        self.move_text = ""

    def add_header(self, key: str, value: str) -> None:
        """
        Add a header to Game.headers

        :param key: The header to be added (usually sourced from the PGN file)
        :param value: The value of the header

        :raises ValueError: If the key has already been added to the Game
        """
        if key in self.headers:
            raise ValueError(
                f"{key} header has already been set on this game - {self.headers[key]}."
            )
        self.headers[key] = value


class GameNode:
    """
    A class representing a node in a game. Useful for
    parsing a PGN game and building a tree of moves
    for the opening book.

    Provides a ``children`` dictionary that maps a move (str) made from
    this GameNode to the GameNode representing the new Board state.

    :param turn: The side to move on this GameNode.
    """

    def __init__(self, turn: str) -> None:
        self.turn: str = turn
        self.children: dict[str:GameNode] = {}  # Maps SAN move strings to GameNode

    def __repr__(self):
        children = ""
        for c in self.children:
            children += " " + c
        return f"<chessengine.GameNode: {self.turn[0]} - {{{children.strip()}}}>"

    def __contains__(self, move: str) -> bool:
        return move in self.children

    def add_child(self, move: str):
        """
        Create a new GameNode and add it as a child to the current node in
        the ``children`` dictionary.

        :param move: The move made to reach the new game node
        :return GameNode: Return the newly created GameNode.
        """
        if move not in self.children:
            turn = "white" if self.turn == "black" else "black"
            new_node = GameNode(turn)
            self.children[move] = new_node

        return self.children[move]

    def get_child(self, move: str):
        """
        Check if ``move`` is a child of this GameNode. If it is,
        return the corresponding GameNode, else raise ValueError.

        :param move: The move to check for
        :return GameNode: Return the GameNode if found
        :raises ValueError: If the move passed is not a child of the GameNode
        """
        if move not in self.children:
            raise ValueError(f"{move} is not a child of the current Game Node.")
        return self.children[move]



import argparse
import sys
from functools import wraps

from chessengine.bitboard import Board


def handle_error(f):
    @wraps(f)
    def wrapper():
        try:
            f()
        except KeyboardInterrupt as e:
            print(f"\nDetected {e.__class__.__name__} {e}.")
            print("Exit.")
            sys.exit(1)

    return wrapper


def prompt_player_side() -> str:
    side = input("Do you want to play white or black (w/b)? - ").lower().strip()
    while not side.startswith(("b", "w")):
        print(
            f'You entered an invalid side - {side}. Enter "w" for white or "b" for black'
        )
        side = input("Do you want to play white or black (w/b)? - ").lower().strip()
    return side


def play(play_with_player: bool) -> None:
    board = Board("black")
    if play_with_player:
        board.play_pvp()
    else:
        player_side = prompt_player_side()

        if player_side.startswith("b"):
            board = Board("white")
        board.play()


def update():
    print("Work in progress.")


@handle_error
def main():
    parser = argparse.ArgumentParser(
        prog=__package__, description="A chess engine written in Python"
    )

    parser.add_argument(
        "-v",
        "--verbose",
        help="set verbose",
        action="store_true",
        required=False,
    )

    subparsers = parser.add_subparsers(help="Action to perform", dest="action")
    parser_play = subparsers.add_parser("play", help="Start a new game.")
    parser_play.add_argument(
        "-p",
        "--player",
        help="Play against another player.",
        required=False,
        action="store_true",
    )

    parser_update = subparsers.add_parser("update", help="Update something.")

    args = parser.parse_args()
    if args.action == "play":
        play(args.player)
    elif args.action == "update":
        update()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# PYTHON_ARGCOMPLETE_OK
"""pytest: unit and functional testing with Python."""
from _pytest import __version__
from _pytest import version_tuple
from _pytest._code import ExceptionInfo
from _pytest.assertion import register_assert_rewrite
from _pytest.cacheprovider import Cache
from _pytest.capture import CaptureFixture
from _pytest.config import cmdline
from _pytest.config import Config
from _pytest.config import console_main
from _pytest.config import ExitCode
from _pytest.config import hookimpl
from _pytest.config import hookspec
from _pytest.config import main
from _pytest.config import PytestPluginManager
from _pytest.config import UsageError
from _pytest.config.argparsing import OptionGroup
from _pytest.config.argparsing import Parser
from _pytest.debugging import pytestPDB as __pytestPDB
from _pytest.fixtures import fixture
from _pytest.fixtures import FixtureLookupError
from _pytest.fixtures import FixtureRequest
from _pytest.fixtures import yield_fixture
from _pytest.freeze_support import freeze_includes
from _pytest.legacypath import TempdirFactory
from _pytest.legacypath import Testdir
from _pytest.logging import LogCaptureFixture
from _pytest.main import Session
from _pytest.mark import Mark
from _pytest.mark import MARK_GEN as mark
from _pytest.mark import MarkDecorator
from _pytest.mark import MarkGenerator
from _pytest.mark import param
from _pytest.monkeypatch import MonkeyPatch
from _pytest.nodes import Collector
from _pytest.nodes import File
from _pytest.nodes import Item
from _pytest.outcomes import exit
from _pytest.outcomes import fail
from _pytest.outcomes import importorskip
from _pytest.outcomes import skip
from _pytest.outcomes import xfail
from _pytest.pytester import HookRecorder
from _pytest.pytester import LineMatcher
from _pytest.pytester import Pytester
from _pytest.pytester import RecordedHookCall
from _pytest.pytester import RunResult
from _pytest.python import Class
from _pytest.python import Function
from _pytest.python import Metafunc
from _pytest.python import Module
from _pytest.python import Package
from _pytest.python_api import approx
from _pytest.python_api import raises
from _pytest.recwarn import deprecated_call
from _pytest.recwarn import WarningsRecorder
from _pytest.recwarn import warns
from _pytest.reports import CollectReport
from _pytest.reports import TestReport
from _pytest.runner import CallInfo
from _pytest.stash import Stash
from _pytest.stash import StashKey
from _pytest.tmpdir import TempPathFactory
from _pytest.warning_types import PytestAssertRewriteWarning
from _pytest.warning_types import PytestCacheWarning
from _pytest.warning_types import PytestCollectionWarning
from _pytest.warning_types import PytestConfigWarning
from _pytest.warning_types import PytestDeprecationWarning
from _pytest.warning_types import PytestExperimentalApiWarning
from _pytest.warning_types import PytestRemovedIn8Warning
from _pytest.warning_types import PytestUnhandledCoroutineWarning
from _pytest.warning_types import PytestUnhandledThreadExceptionWarning
from _pytest.warning_types import PytestUnknownMarkWarning
from _pytest.warning_types import PytestUnraisableExceptionWarning
from _pytest.warning_types import PytestWarning

set_trace = __pytestPDB.set_trace


__all__ = [
    "__version__",
    "approx",
    "Cache",
    "CallInfo",
    "CaptureFixture",
    "Class",
    "cmdline",
    "Collector",
    "CollectReport",
    "Config",
    "console_main",
    "deprecated_call",
    "exit",
    "ExceptionInfo",
    "ExitCode",
    "fail",
    "File",
    "fixture",
    "FixtureLookupError",
    "FixtureRequest",
    "freeze_includes",
    "Function",
    "hookimpl",
    "HookRecorder",
    "hookspec",
    "importorskip",
    "Item",
    "LineMatcher",
    "LogCaptureFixture",
    "main",
    "mark",
    "Mark",
    "MarkDecorator",
    "MarkGenerator",
    "Metafunc",
    "Module",
    "MonkeyPatch",
    "OptionGroup",
    "Package",
    "param",
    "Parser",
    "PytestAssertRewriteWarning",
    "PytestCacheWarning",
    "PytestCollectionWarning",
    "PytestConfigWarning",
    "PytestDeprecationWarning",
    "PytestExperimentalApiWarning",
    "PytestRemovedIn8Warning",
    "Pytester",
    "PytestPluginManager",
    "PytestUnhandledCoroutineWarning",
    "PytestUnhandledThreadExceptionWarning",
    "PytestUnknownMarkWarning",
    "PytestUnraisableExceptionWarning",
    "PytestWarning",
    "raises",
    "RecordedHookCall",
    "register_assert_rewrite",
    "RunResult",
    "Session",
    "set_trace",
    "skip",
    "Stash",
    "StashKey",
    "version_tuple",
    "TempdirFactory",
    "TempPathFactory",
    "Testdir",
    "TestReport",
    "UsageError",
    "WarningsRecorder",
    "warns",
    "xfail",
    "yield_fixture",
]


def __getattr__(name: str) -> object:
    if name == "Instance":
        # The import emits a deprecation warning.
        from _pytest.python import Instance

        return Instance
    raise AttributeError(f"module {__name__} has no attribute {name}")



import unittest
from unittest.mock import patch
from chessengine.bitboard import Board
from typing import Optional, List


class TestBoard(unittest.TestCase):
    def test_get_bitboard(self):
        board = Board("white")
        self.assertEqual(board.get_bitboard("black", "kings"), board.black_kings)
        self.assertEqual(board.get_bitboard("white", "kings"), board.white_kings)
        self.assertEqual(board.get_bitboard("white", "rooks"), board.white_rooks)
        self.assertEqual(board.get_bitboard("black", "rooks"), board.black_rooks)

    def test_get_self_piece_bitboard(self):
        board = Board("white")
        self.assertEqual(board.get_self_piece_bitboard("pawns"), board.white_pawns)

    def test_set_piece_bitboard(self):
        board = Board("white")
        board.set_bitboard("white", "pawns", 0)
        self.assertEqual(board.white_pawns, 0)

    def test_identify_piece_at(self):
        board = Board("white")
        self.assertEqual(
            board.identify_piece_at(2**8), ("white", "pawns", board.white_pawns)
        )
        self.assertEqual(
            board.identify_piece_at(2**7), ("white", "rooks", board.white_rooks)
        )
        self.assertEqual(board.identify_piece_at(2**20), (None, None, None))

    def test_move(self):
        board = Board("white")
        board.move(2**8, 2**16)
        self.assertEqual(board.white_pawns, 0b11111111 << 9)
        board.move(2**9, 2**17)
        self.assertEqual(board.white_pawns, 0b11111111 << 10)

    def test_invalid_move(self):
        board = Board("white")
        with self.assertRaises(ValueError):
            board.move(1, 2**8)

    def test_fen_representation(self):
        return  # Skip this test, FEN representation is deprecated
        board = Board("white")
        self.assertEqual(
            board.FEN,
            "rnbqkbnr/pppppppp/00000000/00000000/00000000/00000000/PPPPPPPP/RNBQKBNR w KQkq - 0 1",
        )

        board.move(2**8, 2**16)
        self.assertEqual(
            board.FEN,
            "rnbqkbnr/pppppppp/00000000/00000000/00000000/P0000000/0PPPPPPP/RNBQKBNR w KQkq - 0 1",
        )

        board.move(2**16, 2**8)
        self.assertEqual(
            board.FEN,
            "rnbqkbnr/pppppppp/00000000/00000000/00000000/00000000/PPPPPPPP/RNBQKBNR w KQkq - 0 1",
        )

        board.undo_move()
        self.assertEqual(
            board.FEN,
            "rnbqkbnr/pppppppp/00000000/00000000/00000000/P0000000/0PPPPPPP/RNBQKBNR w KQkq - 0 1",
        )

        board.move(2**14, 2**54)
        board.undo_move()
        self.assertEqual(
            board.FEN,
            "rnbqkbnr/pppppppp/00000000/00000000/00000000/P0000000/0PPPPPPP/RNBQKBNR w KQkq - 0 1",
        )

    @patch("chessengine.bitboard.Board.move_san")
    @patch("chessengine.bitboard.Board.get_input")
    def test_handle_player_move(self, mock_input, mock_move_san):
        mock_input.return_value = "a2a3"

        board = Board("white")
        move, lines_added, move_undone = board.handle_player_move("white", "")

        self.assertEqual(move, "a2a3")
        self.assertEqual(lines_added, 2)
        self.assertFalse(move_undone)
        mock_move_san.assert_called_with(move=move, side="white")

    @patch("chessengine.bitboard.Board.undo_move")
    @patch("chessengine.bitboard.Board.move_san")
    @patch("chessengine.bitboard.Board.get_input")
    def test_handle_player_move__undo(self, mock_input, mock_move_san, mock_undo):
        mock_input.return_value = "u"

        board = Board("white")
        move, lines_added, move_undone = board.handle_player_move("white", "")

        self.assertEqual(move, "u")
        self.assertEqual(lines_added, 2)
        self.assertTrue(move_undone)
        mock_move_san.assert_not_called()
        mock_undo.assert_called()

    @patch("chessengine.bitboard.Board.undo_move")
    @patch("chessengine.bitboard.Board.move_san")
    @patch("chessengine.bitboard.Board.get_input")
    def test_handle_player_move__undo_with_no_moves(
        self, mock_input, mock_move_san, mock_undo
    ):
        mock_input.return_value = "u"
        mock_undo.side_effect = RuntimeError("No moves have been made yet to undo.")

        board = Board("white")
        move, lines_added, move_undone = board.handle_player_move("white", "")

        self.assertEqual(move, "u")
        self.assertEqual(lines_added, 3)
        self.assertTrue(move_undone)
        mock_move_san.assert_not_called()
        mock_undo.assert_called()


if __name__ == "__main__":
    unittest.main()



"""
Utility functions for common bitboard operations.
"""


from typing import List
from math import log2


piece_characters = {
    ("white", "kings"): "K",
    ("white", "queens"): "Q",
    ("white", "rooks"): "R",
    ("white", "bishops"): "B",
    ("white", "knights"): "N",
    ("white", "pawns"): "P",
    ("black", "kings"): "k",
    ("black", "queens"): "q",
    ("black", "rooks"): "r",
    ("black", "bishops"): "b",
    ("black", "knights"): "n",
    ("black", "pawns"): "p",
}


def get_bit_positions(bitboard: int) -> List[int]:
    """
    Returns a list of positions in the bitboard which have a 1.
    For e.g. - 1001100 returns [100, 1000, 1000000]

    :param bitboard: A bitboard.
    """
    positions = []
    if bitboard == 0:
        return positions

    _ = ~bitboard + 1
    positions.append(bitboard & _)
    positions += get_bit_positions(bitboard & ~_)
    return positions


def get_rank(position: int, log: bool = False) -> int:
    """
    Returns the rank of a position. position either has to be a
    power of 2, or log has to be True.
    Returns the rank which is in range [1, 8]

    :param position: The position whose rank is to be returned. See :ref:`position_representation`
    :param log: If True, position will not be assumed to be a power of 2, but an index on the
        chess board.

    .. note:
        The following calls are equivalent -
        ``get_rank(2**8, False)``
        ``get_rank(8, True)``
    """
    if not log:
        position = log2(position)
    return int(position / 8) + 1


def get_file(position: int, log: bool = False) -> int:
    """
    Returns the file of a position. position either has to be a
    power of 2, or log has to be True.
    Returns the file which is in range [1, 8]

    :param position: The position whose file is to be returned. See :ref:`position_representation`
    :param log: If True, position will not be assumed to be a power of 2, but an index on the
        chess board.

    .. note:
        The following calls are equivalent -
        ``get_file(2**8, False)``
        ``get_file(8, True)``
    """
    if not log:
        position = log2(position)
    return int((position % 8) + 1)


def lsb_pos(board: int) -> int:
    """
    Clears all but the rightmost set bit on the board.
    i.e. - Returns 0000100 for 1010100

    :param board: A bitboard
    """
    return board & ~(board - 1)


def clear_lines(n: int) -> None:
    """
    Clears the last n lines printed so we can print there again

    :param n: The number of lines to clear
    """
    LINE_UP = "\033[1A"
    LINE_CLEAR = "\x1b[2K"
    for i in range(n):
        print(LINE_UP, end=LINE_CLEAR)


def change_turn(side_to_move: str) -> str:
    """
    Returns "black" if side_to_move is "white", and "white" otherwise.

    :param side_to_move: The current side
    """
    if side_to_move == "white":
        return "black"
    return "white"


def get_input(prompt: str) -> str:
    """
    Simple wrapper for input to allow testing
    """
    return input(prompt).strip()



from chessengine.bitboard import Board


if __name__ == "__main__":
    play_against = (
        input(
            'Do you want to play against the computer or another player?\nEnter "c" for computer, "p" for player - '
        )
        .strip()
        .lower()
    )
    while play_against not in ["p", "c"]:
        play_against = input(
            'You entered an invalid character. Enter "p" to play against another player, and "c" to play against the computer - '
        )

    if play_against == "c":
        player_side = input("Do you want to play white or black (w/b)? - ")
        while not player_side.lower().strip().startswith(("b", "w")):
            print(
                f"You entered an invalid side - {player_side}. Enter b for black and w for white"
            )
            player_side = input("Do you want to play white or black (w/b)? - ")
        if player_side.lower().strip().startswith("w"):
            b = Board("black")
        else:
            b = Board("white")
        b.play()
    elif play_against == "p":
        b = Board("black")
        b.play_pvp()



import unittest
from chessengine.bitboard import Board
from chessengine.moves import (
    get_white_pawn_moves,
    get_white_rook_moves,
    get_white_bishop_moves,
    get_white_knight_moves,
    get_white_king_moves,
)


# TODO - Add tests for black pieces
class TestMoves(unittest.TestCase):
    def test_get_white_pawn_moves(self):
        board = Board("white")
        moves = get_white_pawn_moves(board, 2**8, False)
        self.assertEqual(moves, [(2**8, 2**16), (2**8, 2**24)])

        board.move(2**9, 2**16)
        self.assertEqual(get_white_pawn_moves(board, 2**8, False), [])

        board.move(2**8, 2**49)
        self.assertEqual(
            get_white_pawn_moves(board, 2**49, False),
            [(2**49, 2**56), (2**49, 2**58)],
        )

    def test_get_white_rook_moves(self):
        board = Board("white")
        moves = get_white_rook_moves(board, 1)
        self.assertEqual(moves, [])

        board.move(2**8, 2**16)
        moves = get_white_rook_moves(board, 1)
        self.assertEqual(moves, [(1, 2**8)])

        board = Board("white")
        board.move(1, 2**28)
        moves = get_white_rook_moves(board, 2**28)
        self.assertEqual(
            moves,
            [
                (2**28, 2**36),
                (2**28, 2**44),
                (2**28, 2**52),
                (2**28, 2**20),
                (2**28, 2**29),
                (2**28, 2**30),
                (2**28, 2**31),
                (2**28, 2**27),
                (2**28, 2**26),
                (2**28, 2**25),
                (2**28, 2**24),
            ],
        )

    def test_get_white_bishop_moves(self):
        board = Board("white")
        moves = get_white_bishop_moves(board, 4)
        self.assertEqual(moves, [])

        board.move(4, 2**18)
        moves = get_white_bishop_moves(board, 2**18)
        self.assertEqual(
            moves,
            [
                (2**18, 2**27),
                (2**18, 2**36),
                (2**18, 2**45),
                (2**18, 2**54),
                (2**18, 2**25),
                (2**18, 2**32),
            ],
        )

    def test_get_white_knight_moves(self):
        board = Board("white")
        moves = get_white_knight_moves(board, 2)
        self.assertEqual(moves, [(2, 2**16), (2, 2**18)])

        board.move(2, 2**26)
        moves = get_white_knight_moves(board, 2**26)
        self.assertEqual(
            moves,
            [
                (2**26, 2**16),
                (2**26, 2**20),
                (2**26, 2**41),
                (2**26, 2**43),
                (2**26, 2**32),
                (2**26, 2**36),
            ],
        )

    def test_get_white_king_moves(self):
        board = Board("white")
        moves = get_white_king_moves(board, 16)
        self.assertEqual(moves, [])

        board.move(2**4, 2**28)
        moves = get_white_king_moves(board, 2**28)
        self.assertEqual(
            moves,
            [
                (2**28, 2**20),
                (2**28, 2**19),
                (2**28, 2**21),
                (2**28, 2**36),
                (2**28, 2**35),
                (2**28, 2**37),
                (2**28, 2**27),
                (2**28, 2**29),
            ],
        )

        board.move(2**11, 2**19)
        moves = get_white_king_moves(board, 2**28)
        self.assertEqual(
            moves,
            [
                (2**28, 2**20),
                (2**28, 2**21),
                (2**28, 2**36),
                (2**28, 2**35),
                (2**28, 2**37),
                (2**28, 2**27),
                (2**28, 2**29),
            ],
        )


if __name__ == "__main__":
    unittest.main()



"""
The PGN parser class
"""


import re
from chessengine.pgn.node import GameNode, Game

# groups()[0] = The piece moved, one of {K, Q, N, R, B}, None if a pawn was moved
# groups()[1] = The file the piece was moved from. Present to resolve ambiguity, if any
# groups()[2] = The rank the piece was moved from. Present to resolve ambiguity, if any
# groups()[3] = The square the piece was moved to. Always non-null
# groups()[4] = The piece a pawn was promoted to, if a pawn reached the last rank
SAN_MOVE_REGEX = re.compile(
    "([KQNRB\u2654\u2655\u2656\u2657\u2658\u2659\u265A\u265B\u265C\u265D\u265E\u265F])?([a-h])?([1-8])?x?([a-h][1-8])=?([QNRB])?"
)
MOVE_TEXT_COMMENT_REGEX = re.compile(
    r"\s*{.*?}\s*"
)  # TODO - add support for removing line comments
MOVE_TEXT_MOVE_REGEX = re.compile(
    r"(\d+)\.\s*([A-Za-z0-9\-+=]+)\s+([A-Za-z0-9\-+=]+)\s*(1-0|0-1|1/2-1/2)?\s*"
)


class PGNParser:
    """
    A parser for parsing PGN files and constructing a tree of GameNodes

    :param pgn_files: A list of PGN files passed as strings, path-like objects, or file-like objects
    :ivar root_node: The root node of the Game tree that the opening book will start from (always corresponds to a new chess board)
    :ivar games: A list of games parsed by the parser
    """

    def __init__(self, pgn_files: list[str] = None) -> None:
        self.pgn_files = pgn_files  # Path to the pgn_file to parse OR file object
        self.root_node: GameNode = GameNode("white")
        self.current_node = self.root_node
        self.current_game = None
        self.games: list[Game] = []

    def parse(self, pgn_file=None):
        """
        Parse the given pgn file if pgn_file is not None.
        Otherwise, parse all the pgn files in self.pgn_files.

        :param pgn_file: a path to a PGN file as a string, path-like object, or file-like object
        """
        files_to_parse = [pgn_file] if pgn_file is not None else self.pgn_files
        for file in files_to_parse:
            try:
                self._parse(file)
            except (TypeError, AttributeError):
                with open(file, mode="r", errors="replace") as opened_pgn_file:
                    self._parse(opened_pgn_file)

    def _parse(self, pgn_file):
        """
        Parses a PGN file and builds a tree of GameNodes
        """
        lines = pgn_file.readlines()
        new_game = True
        move_text = ""
        for line in lines:
            if not line:
                continue
            if line.startswith("["):
                if move_text:
                    # Completely parses the move text of a game. Ends that game.
                    self._parse_move_text(move_text)
                    new_game = True
                if new_game:
                    # New game starts here. Reset current node.
                    self.current_game = Game(self.root_node)
                    self.games.append(self.current_game)

                    self.current_node = self.root_node
                    move_text = ""
                    new_game = False
                self._parse_header(line)
            else:
                # This is the move text
                new_game = False
                move_text += " " + line.strip()
        self._parse_move_text(move_text)

    def _parse_header(self, header_string: str):
        """
        Parses a header string in a PGN file and sets the
        header on the current game
        """
        header_string = header_string.strip()[1:][:-1]
        _ = header_string.split()
        key = _[0]
        value = " ".join(_[1:])
        if value.startswith('"') and value.endswith('"'):
            value = value[1:][:-1]
        self.current_game.add_header(key, value)

    def _parse_move_text(self, move_text: str):
        if "{" in move_text:
            # Remove all comments from the move text
            move_text = MOVE_TEXT_COMMENT_REGEX.sub(" ", move_text).strip()
        self.current_game.move_text = move_text

        ply_number = 0
        move_list = MOVE_TEXT_MOVE_REGEX.findall(move_text.strip())
        last_move = move_list.pop()
        for _, white_move, black_move, _ in move_list:
            self.current_node = self.current_node.add_child(white_move)
            ply_number += 1
            self.current_node = self.current_node.add_child(black_move)
            ply_number += 1

        if last_move[2] in {"1-0", "0-1", "1/2-1/2"}:
            self.current_game.result = last_move[2]
        else:
            self.current_node = self.current_node.add_child(last_move[2])
        if last_move[3] in {"1-0", "0-1", "1/2-1/2"}:
            self.current_game.result = last_move[3]
